@@ -4,7 +4,7 @@ import pytest
 from dowhy.gcm.density_estimators import GaussianMixtureDensityEstimator, KernelDensityEstimator1D
 
 
-def test_gaussian_mixture_density_estimator():
+def test_when_fit_and_evaluate_gaussian_mixture_density_estimator_then_behaves_as_expected():
     test_data = np.array([[0, 1], [0, 2], [1, 0], [2, 3]])
 
     density_estimator_model = GaussianMixtureDensityEstimator()
@@ -17,7 +17,7 @@ def test_gaussian_mixture_density_estimator():
     assert len(results) == 4
 
 
-def test_kernel_based_density_estimator_1d():
+def test_when_fit_and_evaluate_kernel_based_density_estimator_1d_then_behaves_as_expected():
     test_data = np.array([[0, 1], [0, 2], [1, 0], [2, 3]])
 
     density_estimator_model = KernelDensityEstimator1D()
